-- Pylon: a foundation library; version 0.82(beta); 1998-01-04

-- 
-- Copyright (c) 1997 Franck Arnaud, England, E.U.
-- 
-- Permission is hereby granted, without written agreement and without
-- licence or royalty fees, to use, copy, modify, or distribute this source
-- code or binary programs using it or part of it, subject to the following
-- condition:
-- 
-- The above copyright notice and the following disclaimer must appear in
-- all copies or substantial portions of this source code. (Binary programs
-- derived from the source code can be distributed without restriction or
-- mention of the origin of the source.)
-- 
-- THIS SOFTWARE IS PROVIDED BY FRANCK ARNAUD AND THE AUTHORS ``AS IS'' AND
-- ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
-- IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
-- ARE DISCLAIMED. IN NO EVENT SHALL THE AUTHORS BE LIABLE TO ANY PARTY FOR 
-- ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL 
-- DAMAGES ARISING IN ANY WAY OUT OF THE USE OF THIS SOURCE CODE.
-- 

indexing
	cluster: Pylon, Container
	description: "Set of hashable values "
	interface: client
	implementation: "Automatically resized hash table (indexed list of linked lists)"
	warning: "Hash code must not change after item has been added to container"

class P_HASH_SET[G -> HASHABLE]

inherit
	P_SET[G]
		end

creation
	make

feature -- Creation

	make, reset is
			-- Create set.
		do
			count := 0
			init(Default_hashsize)
		end

feature {NONE} -- Constants

	Default_hashsize : INTEGER is 29
			-- Default size of hast table
			-- Should a prime number!

	Grow_entries : INTEGER is 10
			-- Average number of entries per key before growth.

	Shrink_entries : INTEGER is 2
			-- Average number of entries per key before shrink.

	Resize_factor : INTEGER is 2
			-- Resize factor when growing/shrinking.

	Prime_interval : INTEGER is 25
			-- Interval to search for a prime number.

feature {NONE} -- Internal attributes

	found_result : BOOLEAN
			-- Result of last search.
	hashlist : P_LINKED_LIST[G]
			-- Current list.

feature {P_HASH_SET} -- Exported for copy
	
	hashsize : INTEGER
			-- Size of the hashtable.
	hashtable : ARRAY[P_LINKED_LIST[G]]
			-- Hash table (sequence of linked lists).

feature
	
	count : INTEGER

	is_writable : BOOLEAN is true

feature
	
	item : G is
			-- Current item.
			-- Time: O(1)
		do
			Result := hashlist.item
		end

feature -- Collection interface

	to_list : P_LIST[G] is
			-- Convert the hash set to a (linked) list.
			-- Time: O(count)
		local
			i : INTEGER
			list : P_LINKED_LIST[G]
		do
			!!list.make
			Result := list

			from
				i := hashtable.lower
			invariant
				bounds: i >= hashtable.lower and i <= hashtable.upper + 1 
			variant
				hashtable.upper - i + 1
			until
				i > hashtable.upper
			loop
				Result.append(hashtable.item(i))
				i := i + 1
			end
		end

feature {NONE} -- Hashtable
	
	hash_to_list(it : G) : P_LINKED_LIST[G] is
		do
			Result := hashtable.item((it.hash_code \\ hashsize) + 1)
		ensure
			valid: Result /= Void
		end

	init(size : INTEGER) is
			-- Init hashtable
		local
			i : INTEGER
			temp_hashlist : P_LINKED_LIST[G]
		do
			-- table
			!!hashtable.make(1, size)
			from
				i := 1
			variant
				size - i + 1
			until
				i > size
			loop
				!!temp_hashlist.make -- create empty list
				hashtable.put(temp_hashlist, i)
				i := i + 1
			end
			
			-- attributes
			hashsize := size
			found_result := false
			hashlist := Void
		end

	
	resize(newsize : INTEGER) is
			-- Resize hashtable (rehash)
		local
			primesize : INTEGER
			prime : P_PRIME
		do
			!!prime

			-- try to find neighbouring prime
			prime.get_prime_interval(newsize, newsize + Prime_interval)
			if prime.last_result then
				primesize := prime.last_prime
			else
				check prime_usually_available: false end
				primesize := newsize
			end

			-- Note: the max distance between two primes < 100 is 8 so 
			-- we should have primesize near enough to newsize so that 
			-- the resizing invariant holds.

			rehash(primesize, hashtable)
		end

	rehash(newsize : INTEGER; old_hashtable : ARRAY[P_LINKED_LIST[G]]) is
		local
			temp_hashlist : P_LINKED_LIST[G]
			it_list : P_ITERATOR[G]
			titem : G
			i : INTEGER
		do
			-- init new hashtable
			init(newsize)
			
			-- copy old to new
			from
				i := 1
			invariant
				table_lower: old_hashtable.lower = 1
				bounds: i <= old_hashtable.upper + 1
			variant
				old_hashtable.upper - i + 1
			until
				i > old_hashtable.upper
			loop
				temp_hashlist := old_hashtable.item(i)
				from
					it_list := temp_hashlist.iterator
					it_list.first
				variant
					it_list.forward_variant
				until
					it_list.outside
				loop
					titem := it_list.item
					hash_to_list(titem).add(titem)
					
					it_list.forth
				end

				-- forth
				i := i + 1 
			end

			default_position
		ensure
			size: hashtable.count = hashsize 
		end

	default_position is
			-- First item found.
		require
			not_empty: not is_empty
		local
			i : INTEGER
			it : P_ITERATOR[G]
		do
			from
				i := 1
				hashlist := Void
			variant
				hashtable.count - i + 1
			until
				hashlist /= Void or i > hashtable.count
			loop
				if not hashtable.item(i).is_empty then
					hashlist := hashtable.item(i)
					-- hashlist.first
					it := hashlist.iterator
					it.first
					it.synchronize_container
					it.stop
				end
				i := i + 1
			end
		ensure
			done: hashlist /= Void
		end

feature {NONE} -- Resizing
	
	grow is
		do
			if (count // hashsize) > Grow_entries then
				resize(hashsize * Resize_factor)
			end
		end

	shrink is
		do
			if hashsize > Default_hashsize then
				if (count // hashsize) < Shrink_entries then
					resize(hashsize // Resize_factor)
				end
			end
		end

feature {NONE}
	
	found_success(iter : P_ITERATOR[G]) is
		do
			found_result := true
			iter.synchronize_container
			iter.stop
		end

feature
	
	search(x : G) is
			-- Search item in container (compare using is_equal).
			-- Time: O(1)
		local
			iter : P_ITERATOR[G]
			curlist : P_LINKED_LIST[G]
		do
			curlist := hash_to_list(x)
			found_result := false

			from
				iter := curlist.iterator
				iter.first
			variant
				iter.forward_variant
			until
				iter.outside
			loop
				if (iter.item).is_equal(x) then
					found_success(iter)
					hashlist := curlist
				else
					iter.forth
				end
			end
		end


	found : BOOLEAN is
			-- Result of last search.
			-- Time: O(1), attribute
		do
			Result := found_result
		end

feature
	
	has(it : G) : BOOLEAN is
			-- Is the item in the container (side-effect free `search')
			-- Time: O(1)
		local
			curlist : P_LINKED_LIST[G]
			iter : P_ITERATOR[G]
		do
			curlist := hash_to_list(it)
			from
				iter := curlist.iterator
				iter.first
			variant
				iter.forward_variant
			until
				iter.outside
			loop
				if (iter.item).is_equal(it) then
					Result := true
					iter.stop
				else
					iter.forth
				end
			end
		end

feature
	
	add(it : G) is
			-- Add item to container.
			-- Time: O(1), amortised
		do
			hashlist := hash_to_list(it)
			hashlist.add(it)
			count := count + 1
			-- resize 
			grow
		ensure then
			count: count = old count + 1
		end

	remove is
			-- Remove current item.
			-- Time: 0(1), amortised
		do
			hashlist.remove
			count := count - 1
			if count > 0 then
				default_position
			else
				hashlist := Void
			end
			-- resize
			shrink
		ensure then
			count: count = old count - 1
			done: -- not has(old item)
		end

feature -- General
	
	copy(other : like Current) is
			-- Copy set (same items).
			-- Time: 0(count)
		do
			rehash(other.hashsize, other.hashtable)
			count := other.count
		end

	is_equal(other : like Current) : BOOLEAN is
			-- Other container equal?
			-- Time: O(count)
		local
			it : P_ITERATOR[G]
			i : INTEGER
		do
			if other.count = count then
				Result := true

				from
					i := 1
				variant
					hashsize - i + 1
				until
					Result = false or else i > hashsize
				loop
					from
						it := hashtable.item(i).iterator
						it.first
					variant
						it.forward_variant
					until
						it.outside
					loop
						if not other.has(it.item) then
							Result := false
							it.stop
						else
							it.forth
						end
					end -- inloop
					
					-- forth
					i := i + 1
				end -- exloop
			end
		ensure then
			done_count: Result implies (count = other.count)
			done: -- Result implies (for_each item other.has(item))
		end
	
invariant
	resize_up: (count // hashsize) <= (Grow_entries + 1)
	resize_down: (not is_empty and then hashsize > Default_hashsize) implies
						((count // hashsize) >= (Shrink_entries - 1))
	has_hashtable: hashtable /= Void
	hashsize: hashsize = hashtable.count
	current_item: not is_empty implies hashlist /= Void
	empty: is_empty = (hashlist = Void)

end

